#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
#include <algorithm>

using rvec = std::vector<double>;

double fTest(double t, double y) {
  return y*y*t;
}

double exactSol(double t, double y0) {
  return 2*y0/(2-t*t*y0);
}

double initCond(double x) {
  return -sin(M_PI/2.*x);
}

void oneStepAdamsBashforth2(rvec& yNP1, rvec& yN, rvec& yNM1, double (*f) (double, double), double dt, double tN) {

  for (int i=0; i<yNP1.size(); i++)
    yNP1[i] = yN[i] + dt*1.5*f(tN,yN[i]) - 0.5*dt*f(tN-dt,yNM1[i]); 

};

void oneStepEuler(rvec& yNP1, rvec& yN, double (*f) (double, double), double dt, double tN) {

  for (int i=0; i<yNP1.size(); i++)
    yNP1[i] = yN[i] + dt*f(tN,yN[i]); 

};

int main() {

  int N = 1000;
  // Data vector at different step
  rvec y0(N),y1(N),y2(N);

  // Vectors of time steps and associated errors
  std::vector<double> dts = {0.01,0.005,0.002,0.001};
  std::vector<double> errs;

  // A file in which to put the results
  std::ofstream oFile("results.dat");

  for (double dt: dts) {

    // "Simulation" times
    double t      = 0.;
    double tFinal = 1.;

    // Initial condition
    for (int i=0; i<N; i++)
      y0[i] = initCond(i/(double)(N-1));

    // First step is performed with Euler scheme
    oneStepEuler(y1,y0,&fTest,dt,t);
    t += dt;
 
    // These scalars will be used in the loop to compute the L2 norms
    // of the solution and the difference with approx at all times
    double normSol(0.e0),normDiff(0.e0);

    // Pointers on data vector for faster swapping
    std::vector<rvec*> yN(3);
    yN[2] = &y2;
    yN[1] = &y1;
    yN[0] = &y0;
 
    // Now we use Adams-Bashforth method for all remaining steps
    while (t<tFinal) {
      // Compute next step
      oneStepAdamsBashforth2(*(yN[2]),*(yN[1]),*(yN[0]),&fTest,dt,t);
      t += dt;

      // Add contribution to norms for error computation
      for (int i=0; i<N; i++) {
        double sol = exactSol(t,initCond(i/(double)(N-1)));
        normSol  += pow(sol,2);
        normDiff += pow(sol-(*yN[2])[i],2);
      }
            
      // Shift pointers
      std::rotate(yN.begin(),yN.begin()+1,yN.end());
    }
 
    errs.push_back(sqrt(normDiff/normSol));
    std::cout << "dt: \t" << dt << " \t error: " << errs.back() << std::endl;
    oFile << dt << " \t" << errs.back() << std::endl;

  }

  return 0;

}

